#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gmp.h"
#include <search.h>
#include <time.h>
#include <math.h>
#ifndef WIN32
#include <sys/resource.h>
#include <sys/times.h>
#include <unistd.h>
#endif

#define n 0
#define TRUE 1
#define FALSE 0
#define querylen 20
#define MAX 1000001
#define PN 16757
#define NUM 200


#ifndef WIN32
double  userTime_read, sysTime_read;
double  userTime_sort, sysTime_sort;
double  userTime_rand, sysTime_rand;
double  userTime_write, sysTime_write;
double  userTime_overall, sysTime_overall;
struct rusage myTime_program_start, myTime_read_end,myTime_rand_end, myTime_sort_end, myTime_write_end;
#endif


#ifndef WIN32
void calculateExecutionTime(struct rusage *myTimeStart, struct rusage *myTimeEnd, double *userTime, double *sysTime)
{
	(*userTime) =
		((double) (myTimeEnd->ru_utime.tv_sec  - myTimeStart->ru_utime.tv_sec)) +
		((double) (myTimeEnd->ru_utime.tv_usec - myTimeStart->ru_utime.tv_usec)) * 1e-6;
	(*sysTime) =
		((double) (myTimeEnd->ru_stime.tv_sec  - myTimeStart->ru_stime.tv_sec)) +
		((double) (myTimeEnd->ru_stime.tv_usec - myTimeStart->ru_stime.tv_usec)) * 1e-6;

}
#endif

int alen;

int dupp;


struct statistics_s
{
	int noOfptsprocessed;
};

typedef struct statistics_s statistics;
// globl variable
statistics *stat;

statistics *statistics_new()
{
	statistics *returnValue;

	returnValue = (statistics *) malloc(sizeof(statistics));
	memset(returnValue, 0, sizeof(statistics));

	return returnValue;
}

void printTimeFile(char *d)
{
	FILE *filePtr;
	char fn[1024];
	//////////printf("%s\n",fn);
	filePtr = (FILE *) fopen(d, "w");
	if (filePtr == NULL)
	{
		printf("File %s cannot be opened!\n", d);
		exit(0);
	}

#ifndef WIN32
 	// output execution time
	calculateExecutionTime(&myTime_program_start, &myTime_read_end, &userTime_read, &sysTime_read);
	calculateExecutionTime(&myTime_read_end, &myTime_rand_end, &userTime_rand, &sysTime_rand);
	calculateExecutionTime(&myTime_rand_end, &myTime_sort_end, &userTime_sort, &sysTime_sort);
	calculateExecutionTime(&myTime_sort_end, &myTime_write_end, &userTime_write, &sysTime_write);
	calculateExecutionTime(&myTime_program_start, &myTime_write_end, &userTime_overall, &sysTime_overall);

	// reading time
	fprintf(filePtr, "%.10f,", userTime_rand+sysTime_rand);

	// sorting time
//	fprintf(filePtr, "%.10f,", userTime_sort+sysTime_sort);

	// writing time
	fprintf(filePtr, "%.10f\n", userTime_write+sysTime_write);

	// overall time
//	fprintf(filePtr, "%.10f,%.10f", userTime_overall, sysTime_overall);
#endif

	fclose(filePtr);
}


typedef struct HT
{
    mpz_t vl;
	mpz_t htid;
	//int inv[NUM];
//	int inv_len;
} ht;


typedef struct EN
{
	mpz_t cn;
	mpz_t el;
	mpz_t eu;
} en;


typedef struct EVENT
{
	char **attr;
	int a_len;
} event;



typedef struct NODE
{
	//char name[50];
	event **ev;
	int e_len;
} node;

typedef struct SEQ
{
	node **nd;
	int len;
} seq;

void seq_addnode(seq *sq, node *newnd)
{
	sq->nd[sq->len]=(node *)malloc(sizeof(node));
	sq->nd[sq->len]=newnd;
	sq->len++;
}
void event_init(event *ev)
{
	int i=0;
	ev->a_len=alen;
	ev->attr=(char **)malloc(sizeof(char *)*alen);
	for(i=0;i<alen;i++)
		ev->attr[i]=(char *)malloc(sizeof(char *)*5);
}
void node_addevent(node *nd, event *newev)
{
	int i=0;
	//printf("e1,%s\n",newev->attr[0]);
	nd->ev[nd->e_len]=(event *)malloc(sizeof(event));
	event_init((nd->ev[nd->e_len]));
	for (i=0;i<newev->a_len;i++)
	{
		strcpy(nd->ev[nd->e_len]->attr[i],newev->attr[i]);
	}
	//printf("e2\n");
	nd->ev[nd->e_len]->a_len=newev->a_len;
	//printf("e3\n");
	nd->e_len++;
}

void seq_init(seq *sq, int length)
{
	sq->nd=(node **)malloc(sizeof(node *)*length);
	sq->len=0;
}
void node_init(node *nd)
{
	nd->e_len=0;
	nd->ev=(event **)malloc(sizeof(event *)*400);
}



void print_event(event *et)
{
	int i=0;
	printf("et_len:%d\n",et->a_len);
	for(i=0;i<et->a_len;i++)
		printf("%s,",et->attr[i]);
	printf("et done!\n");
}

void print_node(node *nd)
{
	int i=0;
	printf("nd_len:%d\n",nd->e_len);
	for(i=0;i<nd->e_len;i++)
		print_event(nd->ev[i]);
	printf("node done!\n");
}

void print_seq(seq *sq)
{
	int i=0;
	for (i=0;i<sq->len;i++)
		print_node(sq->nd[i]);
	printf("sq done!\n");
}

void split(char * str, char * split_e, event * newet, int alen)
{
	int ct=0;

	str=strtok(str,split_e);
	while(str!=NULL)
	{
		strcpy(newet->attr[ct],str);
		str=strtok(NULL,split_e);
		ct++;
	}
	newet->a_len=alen;

}
void ReadFile(char * fn, int ***data, int alen, int *len, int sqlen)
{
	FILE *fp;
	char *str,st1[8192];
	int i=0;
	int j=0;
	int k=0;
	char temp[1024];
	int line=0;

	if((fp=fopen(fn,"r"))==NULL)
	{
		printf("Cannot open file %s!\n", fn);
		exit(0);
	}


	for(i=0;i<sqlen;i++)
	{
		fscanf(fp,"%s",temp);
		len[i]=atoi(temp);
		data[i]=(int **)malloc(sizeof(int *)*len[i]);
		for(j=0;j<len[i];j++)
		{
			data[i][j]=(int *)malloc(sizeof(int)*alen);
			for(k=0;k<alen;k++)
			{
				fscanf(fp,"%s",temp);
				data[i][j][k]=atoi(temp);
			}
		}

	}
}


void merge(int *candi, int *clen, int *m, int mlen)
{
	int i,j,flag,ct,pt;
	j=0;
	ct=0;
	int newcandi[MAX];
	int newclen=0;

	if(*clen>0)
	{

	i=0;
	j=0;

	//printf("%d,%d\n",candi->inv_len,m->inv_len);
	for(;i<*clen && j<mlen;)
	{
		if(candi[i]==m[j])
		{
			newcandi[newclen]=candi[i];
			newclen++;
			i++;
			j++;
		}
		else if( candi[i]<m[j])
		{
			i++;
		}
		else if( candi[i] >m[j])
		{
			j++;
		}
	}
	for(i=0;i<newclen;i++)
		candi[i]=newcandi[i];
	*clen=newclen;

	}
	else
	{
		for(i=0;i<mlen;i++)
			candi[i]=m[i];
		*clen=mlen;
	}

}

void EC(mpz_t **ens, int ***data, mpz_t *pdata, int **new_inv, int *new_inv_len, int etlen, int i, int *inv_size, mpz_t *oldprime, mpz_t *newprime, int alen)
{
	mpz_t CN,EL,EU,*N,*prime,tp,tp1,tp2,*c,tp3;
	int nprime=0;
	int *a,*b;
	int j,k,l;
	int ct=0;
	//int alen=6;
	int maxlen;

	maxlen=etlen*(alen-2);

	//printf("%d\n",maxlen);

	prime = (mpz_t *)malloc(sizeof(mpz_t)*(alen-2));
	N = (mpz_t *)malloc(sizeof(mpz_t)*maxlen);
	c = (mpz_t *)malloc(sizeof(mpz_t)*maxlen);
	a = (int *)malloc(sizeof(int)*maxlen);
	b = (int *)malloc(sizeof(int)*maxlen);
	//printf("%d\n",alen);
	//printf("sq:%d-----------------AA\n",i);
		mpz_init_set_str(CN,"1",10);
		mpz_init_set_str(EL,"0",10);
		mpz_init_set_str(EU,"0",10);
		mpz_init_set_str(tp1,"1",10);
		mpz_init_set_str(tp,"0",10);
		mpz_init_set_str(tp2,"1",10);
		for(j=0;j<alen-2;j++)
		{
			mpz_init(prime[j]);
		}
		for(j=0;j<etlen*(alen-2);j++)
		{
			mpz_init(N[j]);
			mpz_init(c[j]);
			a[j]=0;
			b[j]=0;
		}
		for(j=0;j<etlen;j++)
		{
			for(k=0;k<alen-2;k++)
			{

				if(new_inv_len[data[i][j][k]]>0)//(found_item=hsearch(item,FIND))!=NULL)
				{

					if(new_inv[data[i][j][k]][new_inv_len[data[i][j][k]]-1]<i)
					{

						new_inv[data[i][j][k]][new_inv_len[data[i][j][k]]]=i;
						new_inv_len[data[i][j][k]]++;

					}
					else
					{
						dupp++;
					}

					mpz_set(prime[k], pdata[data[i][j][k]]);

				}
				else
				{

					mpz_nextprime(*newprime, *oldprime);
					mpz_set(pdata[data[i][j][k]],*newprime);
					new_inv[data[i][j][k]][0]=i;//pdata[nprime].inv[0]=i;
					new_inv_len[data[i][j][k]]=1;//pdata[nprime].inv_len=1;
					mpz_set(*oldprime,*newprime);
					mpz_set(prime[k],*newprime);

				}

				if (new_inv_len[data[i][j][k]]==inv_size[data[i][j][k]])
				{

					if(inv_size[data[i][j][k]]<1000)
						inv_size[data[i][j][k]]=inv_size[data[i][j][k]]*2;
					else
						inv_size[data[i][j][k]]=inv_size[data[i][j][k]]+100;
					new_inv[data[i][j][k]]=(int *)realloc(new_inv[data[i][j][k]],sizeof(int)*inv_size[data[i][j][k]]);
				}

			}

			for(l=0;l<alen-2;l++)
			{


				if(mpz_divisible_p(CN,prime[l])==0)
				{

					mpz_set(N[ct],prime[l]);
					a[ct]=data[i][j][alen-2];
					b[ct]=data[i][j][alen-1];
					ct++;
				}
				else
				{
					for(k=0;k<ct;k++)
						if(mpz_cmp(N[k],prime[l])==0)
							break;
					b[k]=data[i][j][alen-1];
				}
				mpz_mul(CN,CN,prime[l]);
			}

		}

		mpz_init_set_str(tp3,"1",10);
		for(k=0;k<ct;k++)
		{
			mpz_mul(tp3,tp3,N[k]);
		}


		for(k=0;k<ct;k++)
		{

			mpz_set_str(tp1,"1",10);
			mpz_set_str(tp2,"1",10);
			mpz_set(tp,N[k]);
			mpz_set(N[k],tp1);
			for(l=0;l<ct;l++)
			{
				mpz_mul(tp1,tp1,N[l]);
			}

			mpz_set(c[k],tp1);
			mpz_set(N[k],tp);
			mpz_gcdext(tp2,tp,tp1,N[k],c[k]);
			mpz_mul(c[k],c[k],tp1);

			mpz_set_si(tp1,(long)a[k]);
			mpz_mul(tp,c[k],tp1);
			mpz_add(EL,EL,tp);

			mpz_set_si(tp1,(long)b[k]);
			mpz_mul(tp,c[k],tp1);
			mpz_add(EU,EU,tp);
		}



		mpz_mod(EL,EL,tp3);
		mpz_mod(EU,EU,tp3);
		mpz_set(ens[i][0],CN);
		mpz_set(ens[i][1],EL);
		mpz_set(ens[i][2],EU);

		mpz_clear(CN);
		mpz_clear(EL);
		mpz_clear(EU);
		mpz_clear(tp);
		mpz_clear(tp1);
		mpz_clear(tp2);
		mpz_clear(tp3);
		for(j=0;j<maxlen;j++)
		{
			mpz_clear(N[j]);
			mpz_clear(c[j]);
		}
		for(j=0;j<alen-2;j++)
			mpz_clear(prime[j]);

		free(N);
		free(c);
		free(prime);
		free(a);
		free(b);

}

int main(int argv, char * argc[])
{
	FILE *fp,*fp1,*fp2,*fp3;
	mpz_t **ens;
	char fn[8192];
	int i=0;
	int j=0;
	int k=0;
	int l=0;
	int sqlen;
	int etlen;
	mpz_t oldprime,newprime,*N,*prime,tp,tp1,tp2,**query,*c,*pdata;
	int *pt1,*pt2;
	char temp[8192],chtemp[8192];
	int ct=0;
	int nprime=0;
	int *a,*b;
	int nquery=0;
	char *ch;
	int **realquery;
	int cap=0;
	int candind[MAX];
	int candilen=0;
	int tempht;
	char foundch[1024];
	int maxlen=0;

	int *qlen;
	int **result;
	int flagu=0;
	int flagl=0;
	int p=0;
	int sum=0;
	char fnd[1024],fnq[1024],fnr[1024],fnm[1024],fnl[1024],fpp[1024];

	int ***data;
	int *len;
	int **new_inv;
	int new_inv_len[PN];
	int inv_size[PN];
	int prune1;
	double pru1,pru2,pru3;

	dupp=0;
	prune1=0;

	pru1=0;
	pru2=0;
	pru3=0;



	if (argv!=2)
	{
		printf("parameter error!\n");
		return 0;
	}

	printf("begin!\n");

#ifndef WIN32
	sprintf(fnd,"./%s/data.txt",argc[1]);
	sprintf(fnq,"./%s/query.txt",argc[1]);
	sprintf(fnr,"./%s/result3.txt",argc[1]);
	sprintf(fnm,"./%s/config",argc[1]);
	sprintf(fnl,"./%s/sqlen.txt",argc[1]);
	sprintf(fpp,"./%s/prune.txt",argc[1]);
#endif

#ifdef WIN32
	sprintf(fnd,"%s\\data.txt",argc[1]);
	sprintf(fnq,"%s\\query.txt",argc[1]);
	sprintf(fnr,"%s\\result3.txt",argc[1]);
	sprintf(fnm,"%s\\config",argc[1]);
	sprintf(fnl,"%s\\sqlen.txt",argc[1]);
#endif
	if((fp=fopen(fnm,"r"))==NULL)
	{
		printf("Cannot open file %s!\n",fnm);
		return (0);
	}


	fgets(temp,8192,fp);
	alen=atoi(temp);
	fclose(fp);


	if((fp=fopen(fnl,"r"))==NULL)
	{
		printf("Cannot open file %s!\n",fnl);
		return (0);
	}

	printf("Read File:\n");

	if((fp2=fopen(fnd,"r"))==NULL)
	{
		printf("Cannot open file %s!\n", fnd);
		exit(0);
	}

	sqlen=MAX; //can be read from "config"

	len=(int *)malloc(sizeof(int)*sqlen);
	data=(int ***)malloc(sizeof(int **)*sqlen);
	new_inv=(int **)malloc(sizeof(int*)*PN);
	for(i=0;i<PN;i++)
	{
		new_inv[i]=(int *)malloc(sizeof(int)*NUM);
		inv_size[i]=NUM;
		new_inv_len[i]=0;
	}


	for(i=0;i<sqlen && !feof(fp) ;i++)
	{

		if (i%10000==0)
			printf("sq%d:-----------------\n",i);
		fscanf(fp,"%s",temp);
		if(!feof(fp))
		{
		len[i]=atoi(temp);

		if (maxlen<len[i])
		{
			maxlen=len[i];
		}
		data[i]=(int **)malloc(sizeof(int *)*len[i]);
		for(j=0;j<len[i];j++)
		{
			data[i][j]=(int *)malloc(sizeof(int)*alen);
			for(k=0;k<alen-2;k++)
			{
				fscanf(fp2,"%s",temp);
				data[i][j][k]=atoi(temp);
			}
			data[i][j][alen-2]=j+1;
			data[i][j][alen-1]=j+1;

		}
		}

	}

	fclose(fp2);
	fclose(fp);
	sqlen=i-1;


	mpz_init_set_str(oldprime,"1009",10);
	mpz_init_set_str(newprime,"1009",10);

	while(mpz_cmp_si(oldprime,(long)maxlen)<0)
	{
		mpz_nextprime(newprime,oldprime);
		mpz_set(oldprime,newprime);
	}

	maxlen=maxlen*(alen-2);

	printf("done!\n");

	#ifndef WIN32
	// start time
	getrusage(RUSAGE_SELF,&myTime_read_end);
	#endif

	printf("Encoding:\n");

	pdata=(mpz_t *)malloc(sizeof(mpz_t)*PN);  //pdata=(ht *)malloc(sizeof(ht)*PN);
	for(i=0;i<PN;i++)
	{
		mpz_init(pdata[i]);
	}
	ens=(mpz_t **)malloc(sizeof(mpz_t *)*sqlen);

	for(i=0;i<sqlen;i++)
		ens[i]=(mpz_t *)malloc(sizeof(mpz_t)*3);



	for(i=0;i<sqlen;i++)
	{
		if(i%10000==0)
			printf("seq %d:-----------\n",i+1);
		EC(ens,data,pdata,new_inv,new_inv_len,len[i],i,inv_size, &oldprime, &newprime, alen);
	}
	printf("done!\n");

	#ifndef WIN32
	// start time
	getrusage(RUSAGE_SELF,&myTime_rand_end);
	#endif

		#ifndef WIN32
	// start time
	getrusage(RUSAGE_SELF,&myTime_sort_end);
	#endif

	printf("Read Queries:\n");
	if((fp=fopen(fnq,"r"))==NULL)
	{
		printf("Cannot open file!\n");
		return (0);
	}

	while(!feof(fp))
	{
		fgets(temp,8192,fp);
		nquery++;
	}
	nquery--;
	printf("nquery:%d\n",nquery);
	fclose(fp);

	if((fp=fopen(fnq,"r"))==NULL)
	{
		printf("Cannot open file!\n");
		return (0);
	}

	query=(mpz_t **)malloc(sizeof(mpz_t *)*nquery);
	qlen=(int *)malloc(sizeof(int)*nquery);
	result=(int **)malloc(sizeof(int*)*nquery);
	realquery=(int **)malloc(sizeof(int *)*nquery);
	i=j=0;
	for(i=0;i<nquery;i++)
	{
		fscanf(fp, "%s",temp);
		qlen[i]=atoi(temp);
		query[i]=(mpz_t *)malloc(sizeof(mpz_t)*qlen[i]);
		result[i]=(int *)malloc(sizeof(int)*sqlen);
		realquery[i]=(int *)malloc(sizeof(int)*qlen[i]);
		for(j=0;j<qlen[i];j++)
		{
			fscanf(fp, "%s",temp);
			realquery[i][j]=atoi(temp);
			if(mpz_cmp_si(pdata[realquery[i][j]],(long)0)==0)
			{
				result[i][0]=-1;
			}
			else
			{
				mpz_init(query[i][j]);
				mpz_set(query[i][j],pdata[realquery[i][j]]);

			}
		}
	}

	printf("done!\n");
	fclose(fp);


	printf("Execute queries:\n");

	for(i=0;i<nquery;i++)
	{
		prune1=0;
		for(j=0;j<sqlen;j++)
		{
			result[i][j]=-1;
		}


		candilen=0;
		for(j=0;j<MAX;j++)
		{
			candind[j]=0;
		}

		for(j=0;j<qlen[i];j++)
		{

			merge(candind,&candilen,new_inv[realquery[i][j]],new_inv_len[realquery[i][j]]);

		}
		pt1=(int *)malloc(sizeof(int)*qlen[i]);
		pt2=(int *)malloc(sizeof(int)*qlen[i]);


		for(j=0;j<candilen;j++)
		{


			flagu=flagl=0;

			for(k=0;k<qlen[i];k++)
			{
					mpz_set_str(tp1,"0",10);
					mpz_set_str(tp2,"0",10);
					mpz_mod(tp1,ens[candind[j]][1],query[i][k]);//ens[j].el
					mpz_mod(tp2,ens[candind[j]][2],query[i][k]);
					pt1[k]=(int)mpz_get_si(tp1)+n;
					pt2[k]=(int)mpz_get_si(tp2)+n;


			}
			if(result[i][candind[j]]==FALSE)
			{
				continue;
			}
			else
			{
				for(k=0;k<qlen[i];k++)
				{
					for(l=0;l<k;l++)
					{
						if(pt2[k]<pt1[l])
						{
							result[i][candind[j]]=FALSE;
							break;
						}

					}

					if(result[i][candind[j]]==FALSE)
					{
						break;
					}
					else
					{
						if(k>0&&pt2[k]>pt2[k-1])
							flagu++;

						if(k>0&&pt1[k]>pt1[k-1])
							flagl++;

					}

				}
				if(result[i][candind[j]]==FALSE)
				{
					continue;
				}
				else if(flagu==qlen[i] || flagl==qlen[i])
				{
					result[i][candind[j]]=TRUE;
				}
				else // compare true sequence
				{
				    k=0;
					l=0;
					while(k<len[candind[j]])// sq.nd[candind.inv[j]]->e_len)
					{
						for(p=0;p<alen-2;p++)
						{
							if(data[candind[j]][k][p]==realquery[i][l])
							{
								l++;
							}
						}
						k++;
					}
					if(l==qlen[i])
						result[i][candind[j]]=TRUE;
					else
						result[i][candind[j]]=FALSE;
				}

			}


		}
		free(pt1);
		free(pt2);
		free(query[i]);
		free(realquery[i]);

	}


	#ifndef WIN32
	// start time
	getrusage(RUSAGE_SELF,&myTime_write_end);
	#endif

	printTimeFile(fnr);

	return 0;
}
